import * as THREE from 'three';
import { debounce, clamp as clamp$1, isEqual, isEmpty, range, isNil, inRange as inRange$1, throttle, floor, last } from 'lodash-es';
import { bbox, polygon, lineIntersect, lineString } from '@turf/turf';

/**
 * 获取一元二次方程的解
 */
function getQuadraticEquationRes(a, b, c) {
    const sqrtRes = Math.sqrt(Math.pow(b, 2) - 4 * a * c);
    return [(-b + sqrtRes) / (2 * a), (-b - sqrtRes) / (2 * a)];
}
/**
 * 角度转弧度
 * !!! 会用于 Web Worker
 * @param degrees
 */
function degToRad(degrees) {
    return (degrees / 180) * Math.PI;
}
/**
 * 弧度转角度
 * !!! 会用于 Web Worker
 * @param degrees
 */
function radToDeg(radian) {
    return (radian / Math.PI) * 180;
}
/**
 * 根据直角三角形的两条边获取斜边长
 * @param a
 * @param b
 * @returns
 */
function hypotenuse(a, b) {
    return Math.sqrt(Math.pow(a, 2) + Math.pow(b, 2));
}
function rectangleIntersect(width, height, deg) {
    const halfWidth = width / 2;
    const halfHeight = height / 2;
    const diagonalDeg = radToDeg(Math.atan(width / height));
    const _deg = deg % 360;
    if (_deg === 0) {
        return [halfWidth, 0];
    }
    if (_deg < diagonalDeg) {
        return [halfWidth + halfHeight * Math.tan(degToRad(_deg)), 0];
    }
    if (_deg === diagonalDeg) {
        return [width, 0];
    }
    if (_deg < 90) {
        return [width, halfHeight - halfWidth * Math.tan(degToRad(90 - _deg))];
    }
    if (_deg === 90) {
        return [width, halfHeight];
    }
    if (_deg < 180 - diagonalDeg) {
        return [width, halfHeight + halfWidth * Math.tan(degToRad(_deg - 90))];
    }
    if (_deg === 180 - diagonalDeg) {
        return [width, height];
    }
    if (_deg < 180) {
        return [halfWidth + halfHeight * Math.tan(degToRad(180 - _deg)), height];
    }
    if (_deg === 180) {
        return [halfWidth, height];
    }
    if (_deg < 180 + diagonalDeg) {
        return [halfWidth - halfHeight * Math.tan(degToRad(_deg % 180)), height];
    }
    if (_deg === 180 + diagonalDeg) {
        return [0, height];
    }
    if (_deg < 270) {
        return [0, halfHeight + halfWidth * Math.tan(degToRad(90 - (_deg % 180)))];
    }
    if (_deg === 270) {
        return [0, halfHeight];
    }
    if (_deg < 360 - diagonalDeg) {
        return [0, halfHeight - halfWidth * Math.tan(degToRad(_deg - 270))];
    }
    if (_deg === 360 - diagonalDeg) {
        return [0, 0];
    }
    if (_deg < 360) {
        return [halfWidth - halfHeight * Math.tan(degToRad(360 - _deg)), 0];
    }
}

/**
 * 获取当前距离上刚好与球体相切的视角
 * @param distance 镜头离圆心的距离
 * @param r 半径
 * @returns
 */
function getTangentFov(distance, r) {
    return radToDeg(Math.asin(r / distance)) * 2;
}
/**
 * 获取球体在镜头中可显示区域的圆心角
 * @param distance 镜头离圆心的距离
 * @param r 半径
 * @param fov 镜头视角
 * @returns
 */
function getDisplayCentralAngle(distance, r, fov) {
    const tangentFov = getTangentFov(distance, r);
    if (fov >= tangentFov) {
        return 180 - tangentFov;
    }
    const halfFov = fov / 2;
    const a = 1 / Math.pow(Math.cos(degToRad(halfFov)), 2);
    const b = -2 * distance;
    const c = Math.pow(distance, 2) - Math.pow(r, 2);
    // 摄像头到弦心的距离
    const distanceFromTheCameraToTheChord = Math.min(...getQuadraticEquationRes(a, b, c));
    const α = radToDeg(2 * Math.acos((distance - distanceFromTheCameraToTheChord) / r));
    return α;
}
function lngLatToVector3(lngLat, radius) {
    const [lng, lat] = lngLat;
    return new THREE.Vector3().setFromSphericalCoords(radius, degToRad(90 - lat), degToRad(lng));
}
function vector3ToLngLat(v3) {
    const spherical = new THREE.Spherical().setFromVector3(v3);
    return sphericalToLngLat(spherical);
}
function sphericalToLngLat(spherical) {
    return [radToDeg(spherical.theta), 90 - radToDeg(spherical.phi)];
}
function getSatelliteUrl(x, y, z) {
    return `https://api.mapbox.com/v4/mapbox.satellite/${z}/${x}/${y}@2x.webp?sku=1015N1AhJztkE&access_token=pk.eyJ1IjoiZGluZ2xlaTIwMjEiLCJhIjoiY2wxbHh1aW54MDl6NDNrcGcwODNtaXNtbSJ9.6G649bdbNApupw2unoY0Yg`;
}

class EarthOrbitControls extends THREE.EventDispatcher {
    camera;
    domElement = document.body;
    earthRadius = 6371;
    /**
     * 垂直方向的视角
     */
    fov = 60;
    tileSize = 512;
    center = [0, 0];
    _distance = 0;
    _zoom = 0;
    bearing = 0;
    pitch = 0;
    // enablePitch = true
    minDistance = this.earthRadius + 0.2;
    maxDistance = this.earthRadius * 4;
    lookAtPosition = new THREE.Vector3(0, 0, 0);
    disposeFuncList = [];
    onEnd = debounce(function () {
        this.dispatchEvent({ type: 'end' });
    }, 250);
    constructor(options) {
        super();
        if (options.domElement)
            this.domElement = options.domElement;
        if (options.center)
            this.center = options.center;
        this.earthRadius = options.earthRadius;
        this.zoom = options.zoom ?? 2;
        if (options.bearing)
            this.bearing = options.bearing;
        const { domElement, distance, center } = this;
        const pixelRatio = domElement.clientWidth / domElement.clientHeight;
        const camera = new THREE.PerspectiveCamera(this.fov, pixelRatio, 0.1, this.earthRadius * 10000);
        camera.position.copy(lngLatToVector3(center, distance));
        this.camera = camera;
        this.lookAt();
        const eventListenerList = [
            ['mousemove', this.onMousemove.bind(this)],
            ['contextmenu', this.onContextmenu.bind(this)],
            ['wheel', this.onMousewheel.bind(this)],
            ['pointerup', this.onPointerup.bind(this)],
        ];
        eventListenerList.forEach(([type, listener]) => {
            this.domElement.addEventListener(type, listener);
            this.disposeFuncList.push(() => this.domElement.removeEventListener(type, listener));
        });
    }
    getDistance() {
        const { earthRadius, fov, tileSize, zoom, domElement } = this;
        const pxLat = 360 / (Math.pow(2, zoom) * tileSize);
        const pxLength = Math.sin(degToRad(pxLat)) * earthRadius;
        const chordLength = pxLength * domElement.clientHeight;
        // 摄像头到弦心的距离
        const distanceFromTheCameraToTheChord = chordLength / 2 / Math.tan(degToRad(fov / 2));
        // x^{2} + distanceFromTheCameraToTheChord * x - earthRadius^{2}
        const a = 1;
        const b = distanceFromTheCameraToTheChord;
        const c = -Math.pow(earthRadius, 2);
        // 弦心到圆心的距离
        const distanceFromTheChordToTheCentre = Math.max(...getQuadraticEquationRes(a, b, c));
        const distance = distanceFromTheCameraToTheChord + distanceFromTheChordToTheCentre;
        return distance;
    }
    getZoom() {
        const { tileSize } = this;
        const zoom = Math.log2(360 / this.getPxDeg() / tileSize);
        return zoom;
    }
    /**
     * 每 1 个像素对应的最小度数
     */
    getPxDeg() {
        const { earthRadius, fov, distance, domElement } = this;
        const centralAngle = getDisplayCentralAngle(distance, earthRadius, fov);
        // 弦心到圆心的距离
        const distanceFromTheChordToTheCentre = Math.cos(degToRad(centralAngle / 2)) * earthRadius;
        // 摄像头到弦心的距离
        const distanceFromTheCameraToTheChord = distance - distanceFromTheChordToTheCentre;
        // 弦长
        const chordLength = Math.tan(degToRad(fov / 2)) * distanceFromTheCameraToTheChord * 2;
        // 每 1 个像素对应的弦长
        const pxLength = chordLength / domElement.clientHeight;
        const pxDeg = radToDeg(Math.asin(pxLength / earthRadius));
        return pxDeg;
    }
    get distance() {
        return this._distance;
    }
    set distance(value) {
        this._distance = value;
        this._zoom = this.getZoom();
    }
    get zoom() {
        return this._zoom;
    }
    set zoom(value) {
        this._zoom = value;
        this._distance = this.getDistance();
    }
    get z() {
        return Math.ceil(this.zoom);
    }
    /**
     * 水平方向的视角
     */
    get fovX() {
        // 邻边
        const adjacent = this.domElement.clientHeight / Math.tan(degToRad(this.fov / 2));
        return radToDeg(Math.atan(this.domElement.clientWidth / adjacent)) * 2;
    }
    /**
     * 根据对角线的像素长度获取对应的视角
     * @param diagonal
     * @returns
     */
    getFov(diagonal) {
        // 邻边
        const adjacent = this.domElement.clientHeight / Math.tan(degToRad(this.fov / 2));
        return radToDeg(Math.atan(diagonal / adjacent)) * 2;
    }
    lookAt() {
        this.camera.lookAt(this.lookAtPosition);
        this.camera.rotateZ(-degToRad(this.bearing));
    }
    onMousemove(e) {
        e.preventDefault();
        if (e.buttons === 0) {
            return;
        }
        const { camera, domElement } = this;
        const isMove = !e.ctrlKey;
        if (isMove) {
            // TODO 增加阻尼效果
            const movementXDeg = (e.movementX / domElement.clientWidth) * 36;
            const movementYDeg = (e.movementY / domElement.clientHeight) * 36;
            camera.position.applyAxisAngle(new THREE.Vector3(0, 1, 0).applyQuaternion(camera.quaternion), -degToRad(movementXDeg));
            const applyXAxisAngle = (vector3) => vector3.applyAxisAngle(new THREE.Vector3(1, 0, 0).applyQuaternion(camera.quaternion), -degToRad(movementYDeg));
            applyXAxisAngle(camera.position);
            applyXAxisAngle(camera.up);
            this.lookAt();
            // TODO 移动时 bearing 也会改变
            this.center = vector3ToLngLat(camera.position);
            this.dispatchEvent({ type: 'move' });
            return;
        }
        const isRotate = Math.abs(e.movementX) > Math.abs(e.movementY);
        if (isRotate) {
            const movementDeg = (e.movementX / domElement.clientWidth) * 360;
            this.bearing -= movementDeg;
            this.lookAt();
            this.dispatchEvent({ type: 'rotate' });
            return;
        }
        // TODO pitch 处理有 bearing 的情况
        const movementDeg = (e.movementY / domElement.clientHeight) * 180;
        this.pitch = clamp$1(this.pitch + movementDeg, 0, 85);
        const negativePosition = camera.position.clone();
        negativePosition.x = -negativePosition.x;
        negativePosition.y = -negativePosition.y;
        negativePosition.z = -negativePosition.z;
        const spherical = new THREE.Spherical().setFromVector3(negativePosition);
        spherical.phi -= degToRad(this.pitch);
        const lookAtPosition = new THREE.Vector3().setFromSpherical(spherical);
        lookAtPosition.x -= negativePosition.x;
        lookAtPosition.y -= negativePosition.y;
        lookAtPosition.z -= negativePosition.z;
        this.lookAtPosition = lookAtPosition;
        this.lookAt();
    }
    onContextmenu(e) {
        e.preventDefault();
    }
    onMousewheel(e) {
        e.preventDefault();
        e.stopPropagation();
        const spherical = new THREE.Spherical().setFromVector3(this.camera.position);
        const movement = 1000 / Math.pow(2, this.zoom);
        if (e.deltaY > 0) {
            spherical.radius = Math.min(spherical.radius + movement, this.maxDistance);
        }
        else {
            spherical.radius = Math.max(spherical.radius - movement, this.minDistance);
        }
        const vector3 = new THREE.Vector3().setFromSpherical(spherical);
        this.camera.position.x = vector3.x;
        this.camera.position.y = vector3.y;
        this.camera.position.z = vector3.z;
        this.distance = spherical.radius;
        this.dispatchEvent({ type: 'zoom' });
        this.onEnd();
    }
    onPointerup(e) {
        e.preventDefault();
        this.dispatchEvent({ type: 'end' });
    }
    dispose() {
        this.disposeFuncList.forEach(func => func());
        this.disposeFuncList = [];
    }
}

/**
 * 获取小数部分
 * @param num
 * @returns
 */
/**
 * @param number
 * @param precision
 * @returns
 */
function round(number, precision) {
    precision =
        precision == null ? 0 : precision >= 0 ? Math.min(precision, 292) : Math.max(precision, -292);
    if (precision) {
        // Shift with exponential notation to avoid floating-point issues.
        // See [MDN](https://mdn.io/round#Examples) for more details.
        let pair = `${number}e`.split('e');
        const value = Math.round(`${pair[0]}e${+pair[1] + precision}`);
        pair = `${value}e`.split('e');
        return +`${pair[0]}e${+pair[1] - precision}`;
    }
    return Math.round(number);
}
/**
 * @param number
 * @param lower
 * @param upper
 * @returns
 */
function clamp(number, lower, upper) {
    return Math.min(Math.max(number, lower), upper);
}
/**
 * @param number
 * @param start
 * @param end
 * @param boundary
 */
function inRange(number, start, end, boundary = '[)') {
    switch (boundary) {
        case '()':
            return number > start && number < end;
        case '[]':
            return number >= start && number <= end;
        case '[)':
            return number >= start && number < end;
        case '(]':
            return number > start && number <= end;
    }
}

const fullBBox = [-180, -90, 180, 90];
/**
 * 前者的边界是否完全包括了后者
 * @param parent
 * @param child
 * @returns
 */
function bboxContains(parent, child) {
    return (parent[0] <= child[0] && parent[1] <= child[1] && parent[2] >= child[2] && parent[3] >= child[3]);
}
/**
 * 两个 line 是否存在重叠部分
 * @param a
 * @param b
 * @returns
 */
function lineOverlap(a, b) {
    return (inRange(b[0], a[0], a[1], '[)') ||
        inRange(b[1], a[0], a[1], '(]') ||
        inRange(a[0], b[0], b[1], '[)') ||
        inRange(a[1], b[0], b[1], '(]'));
}
/**
 * 两个 bbox 是否存在重叠部分
 * @param a
 * @param b
 * @returns
 */
function bboxOverlap(a, b) {
    return lineOverlap([a[0], a[2]], [b[0], b[2]]) && lineOverlap([a[1], a[3]], [b[1], b[3]]);
}
function scale(bbox, scaleValue) {
    const [w, s, e, n] = bbox;
    const lngGap = e - w;
    const latGap = n - s;
    const translateLng = ((scaleValue - 1) / 2) * lngGap;
    const translateLat = ((scaleValue - 1) / 2) * latGap;
    return [w - translateLng, s - translateLat, e + translateLng, n + translateLat];
}
/**
 * 在球体显示时，对纬度超过[90, -90]范围进行处理
 * @param bbox
 * @returns
 */
function latPretreatmentBBox(bbox) {
    let [w, s, e, n] = bbox;
    if (n - s > 180) {
        return fullBBox;
    }
    if (e - w > 360) {
        w = -180;
        e = 180;
    }
    const nGt90 = n > 90;
    const sLtMinus90 = s < -90;
    if (sLtMinus90 && nGt90) {
        w = -180;
        e = 180;
    }
    if (sLtMinus90) {
        n = Math.max(-180 - s, n);
        s = -90;
    }
    if (nGt90) {
        s = Math.min(180 - n, s);
        n = 90;
    }
    return [w, s, e, n];
}
function isFull(bbox) {
    return isEqual(bbox, fullBBox);
}

class LayerManager {
    // readonly canvas = new OffscreenCanvas(1, 1)
    // private readonly ctx = this.canvas.getContext('2d') as OffscreenCanvasRenderingContext2D
    canvas = document.createElement('canvas');
    ctx = this.canvas.getContext('2d');
    layers = [];
    bbox = fullBBox;
    /**
     * 当前实际显示区域的 BBox，只会比 bbox 更小
     */
    displayBBox = fullBBox;
    z = 0;
    onUpdate;
    updateCanvasSize(pxDeg) {
        const { bbox, canvas } = this;
        canvas.width = Math.ceil((bbox[2] - bbox[0]) / pxDeg);
        canvas.height = Math.ceil((bbox[3] - bbox[1]) / pxDeg);
        console.log('width:', canvas.width);
        console.log('height:', canvas.height);
    }
    updateCanvas() {
        const { ctx, canvas } = this;
        ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
        this.layers.sort((v1, v2) => v1.zIndex - v2.zIndex).forEach(layer => {
            layer.imageBitmap && ctx.drawImage(layer.imageBitmap, 0, 0, canvas.width, canvas.height);
        });
        this.onUpdate?.();
    }
    refresh() {
        if (isEmpty(this.layers)) {
            return;
        }
        this.layers.forEach(layer => {
            layer.refresh();
        });
        this.updateCanvas();
    }
    update() {
        this.layers.forEach(layer => {
            layer.update();
        });
    }
    addLayer(layer) {
        layer.layerManager = this;
        this.layers.push(layer);
        layer.refresh();
        this.updateCanvas();
    }
    removeLayer(layer) {
        const index = this.layers.findIndex(l => l === layer);
        this.layers.splice(index, 1);
        layer.dispose();
    }
    dispose() {
        this.layers.forEach(layer => layer.dispose());
    }
}

// https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
class MercatorTile {
    static workerScripts = [clamp, round, degToRad, MercatorTile];
    static maxLat = round(Math.atan(Math.sinh(Math.PI)) * 180 / Math.PI, 12);
    static pointToTile(lng, lat, z) {
        return [
            Math.floor(MercatorTile.lngToX(lng, z)),
            Math.floor(MercatorTile.latToY(lat, z)),
        ];
    }
    static tileToBBox(xyz) {
        const [x, y, z] = xyz;
        const e = MercatorTile.xToLng(x + 1, z);
        const w = MercatorTile.xToLng(x, z);
        const s = MercatorTile.yToLat(y + 1, z);
        const n = MercatorTile.yToLat(y, z);
        return [w, s, e, n];
    }
    /**
     * 将边界 box 转化为 tile 索引的 box
     */
    static bboxToTileIndexBox(bbox, z) {
        const [w, s, e, n] = bbox;
        const [startX, startY] = MercatorTile.pointToTile(w, n, z);
        const [x2, y2] = MercatorTile.pointToTile(e, s, z);
        const endX = e === MercatorTile.xToLng(x2, z) ? x2 : x2 + 1;
        const clampedS = clamp(s, -MercatorTile.maxLat, MercatorTile.maxLat);
        const endY = clampedS === MercatorTile.yToLat(y2, z) ? y2 : y2 + 1;
        return {
            startX,
            startY,
            endX,
            endY
        };
    }
    static lngToX(lng, z) {
        return ((lng + 180) / 360) * Math.pow(2, z);
    }
    static xToLng(x, z) {
        return (360 / Math.pow(2, z)) * x - 180;
    }
    static latToY(lat, z) {
        const sin = Math.sin(degToRad(lat));
        const z2 = Math.pow(2, z);
        const y = z2 * (0.5 - (0.25 * Math.log((1 + sin) / (1 - sin))) / Math.PI);
        return clamp(y, 0, z2);
    }
    static yToLat(y, z) {
        const n = Math.PI - (2 * Math.PI * y) / Math.pow(2, z);
        return round((180 / Math.PI) * Math.atan(0.5 * (Math.exp(n) - Math.exp(-n))), 12);
    }
}

/**
 * eg. multiplyArray([1, 2], [3, 4]) = [[1, 3], [1, 4], [2, 3], [2, 4]]
 * @param xArray
 * @param yArray
 * @param row 是否按行来排列
 * @returns
 */
function multiplyArray(xArray, yArray, row) {
    const arr = [];
    if (row) {
        for (const y of yArray) {
            for (const x of xArray) {
                arr.push([x, y]);
            }
        }
    }
    else {
        for (const x of xArray) {
            for (const y of yArray) {
                arr.push([x, y]);
            }
        }
    }
    return arr;
}
/**
 * 如果传入一个非数组字段，则将其转化为数组
 * @param value
 * @returns
 */
function inflate(value) {
    return Array.isArray(value) ? value : [value];
}

class TileCache {
    data = {};
    getKey(xyz) {
        return xyz.join('.');
    }
    has(xyz) {
        return !!this.data[this.getKey(xyz)];
    }
    get(xyz) {
        return this.data[this.getKey(xyz)];
    }
    set(xyz, value) {
        this.data[this.getKey(xyz)] = value;
    }
    delete(xyz) {
        delete this.data[this.getKey(xyz)];
    }
}

class TileLayer {
    canvas;
    tileSize;
    cache = new TileCache();
    bbox = fullBBox;
    canvasBBox = fullBBox;
    /**
     * 当前实际显示区域的 BBox，只会比 bbox 更小
     */
    displayBBox = fullBBox;
    z = 0;
    onUpdate;
    constructor(tileSize) {
        this.tileSize = tileSize;
    }
    /**
     * tileX 有可能小于 0 或者大于等于 z2，需要对其进行格式化
     * @param tileX
     * @returns
     */
    getFormattedTileX(tileX) {
        const z2 = Math.pow(2, this.z);
        return tileX < 0 ? z2 + tileX : tileX % z2;
    }
    async loadTile(xyz) {
        const promise = fetch(getSatelliteUrl(...xyz)).then(async (res) => {
            const imageBitmap = await createImageBitmap(await res.blob());
            this.cache.set(xyz, imageBitmap);
            return imageBitmap;
        });
        this.cache.set(xyz, promise);
        promise.catch(() => {
            this.cache.delete(xyz);
        });
        return await promise;
    }
    /**
     * 如果当前缩放层级的 tile 没有加载完毕，则依次寻找更低缩放级别的图片占位
     * @param x
     * @param y
     * @param formattedX
     */
    drawPreviewImage(x, y, rect) {
        const { tileSize, cache, z } = this;
        const ctx = this.canvas.getContext('2d');
        const z2 = Math.pow(2, z);
        const formattedX = this.getFormattedTileX(x);
        let cloneX = formattedX;
        let cloneY = y;
        let cloneZ = z;
        while (cloneZ > 0) {
            cloneZ--;
            cloneX = Math.floor(cloneX / 2);
            cloneY = Math.floor(cloneY / 2);
            const imageBitmap = cache.get([cloneX, cloneY, cloneZ]);
            if (imageBitmap instanceof ImageBitmap) {
                const cloneZ2 = Math.pow(2, cloneZ);
                const ratio = z2 / cloneZ2;
                const sw = tileSize / ratio;
                const sh = sw;
                const sx = (formattedX % ratio) * sw;
                const sy = (y % ratio) * sw;
                ctx.clearRect(...rect);
                ctx.drawImage(imageBitmap, sx, sy, sw, sh, ...rect);
                continue;
            }
        }
    }
    draw(update) {
        const { tileSize, cache, bbox, z, displayBBox } = this;
        const ctx = this.canvas.getContext('2d');
        const tileIndexBox = MercatorTile.bboxToTileIndexBox(bbox, z);
        const displayTileIndexBox = MercatorTile.bboxToTileIndexBox(displayBBox, z);
        const drawTileIndexBox = update ? displayTileIndexBox : tileIndexBox;
        const xyArr = multiplyArray(range(drawTileIndexBox.startX, drawTileIndexBox.endX), range(drawTileIndexBox.startY, drawTileIndexBox.endY));
        xyArr.forEach(([x, y]) => {
            const formattedX = this.getFormattedTileX(x);
            const xyz = [formattedX, y, z];
            let value = cache.get(xyz);
            const rect = [(x - tileIndexBox.startX) * tileSize, (y - tileIndexBox.startY) * tileSize, tileSize, tileSize];
            const drawImage = (imageBitmap) => {
                ctx.clearRect(...rect);
                ctx.drawImage(imageBitmap, ...rect);
            };
            if (value instanceof ImageBitmap) {
                drawImage(value);
                return;
            }
            if (isNil(value)) {
                this.drawPreviewImage(x, y, rect);
                const inXRange = inRange$1(x, displayTileIndexBox.startX, displayTileIndexBox.endX);
                const inYRange = inRange$1(y, displayTileIndexBox.startY, displayTileIndexBox.endY);
                if (!inXRange || !inYRange) {
                    return;
                }
                value = this.loadTile(xyz);
            }
            void value.then((imageBitmap) => {
                drawImage(imageBitmap);
                this.onUpdate?.();
            });
        });
        this.onUpdate?.();
    }
    async refresh() {
        const defaultTileXyz = [0, 0, 0];
        if (!this.cache.has(defaultTileXyz)) {
            await this.loadTile(defaultTileXyz);
        }
        const { tileSize, bbox, z } = this;
        const tileIndexBox = MercatorTile.bboxToTileIndexBox(bbox, z);
        const width = (tileIndexBox.endX - tileIndexBox.startX) * tileSize;
        const height = (tileIndexBox.endY - tileIndexBox.startY) * tileSize;
        const canvas = new OffscreenCanvas(width, height);
        // 每次刷新时更新 canvas，防止图片异步加载的情况下 ctx.drawImage 错误
        this.canvas = canvas;
        this.canvasBBox = [
            MercatorTile.xToLng(tileIndexBox.startX, z),
            MercatorTile.yToLat(tileIndexBox.endY, z),
            MercatorTile.xToLng(tileIndexBox.endX, z),
            MercatorTile.yToLat(tileIndexBox.startY, z),
        ];
        this.draw();
    }
    update() {
        this.draw(true);
    }
}

class EarthGeometry extends THREE.BufferGeometry {
    bbox = fullBBox;
    earthRadius;
    constructor(options) {
        super();
        this.earthRadius = options.earthRadius;
        if (options.bbox)
            this.bbox = options.bbox;
        if (!options.delay)
            this.update();
    }
    update() {
        const { bbox, earthRadius } = this;
        const heightSegments = 128;
        const widthSegments = heightSegments * 2;
        const widthPositionCount = widthSegments + 1;
        const heightPositionCount = heightSegments + 1;
        const [w, s, e, n] = bbox;
        const lngGap = (e - w) / widthSegments;
        const latGap = (n - s) / heightSegments;
        const positions = [];
        const uvs = [];
        for (let y = 0; y < heightPositionCount; y++) {
            const lat = s + y * latGap;
            for (let x = 0; x < widthPositionCount; x++) {
                const lng = w + x * lngGap;
                const position = lngLatToVector3([lng, lat], earthRadius);
                positions.push(...position.toArray());
                const uvX = x / widthSegments;
                const uvY = y / heightSegments;
                uvs.push(uvX, uvY);
            }
        }
        this.attributes.position = new THREE.Float32BufferAttribute(new Float32Array(positions), 3);
        this.attributes.uv = new THREE.Float32BufferAttribute(new Float32Array(uvs), 2);
        const indexArr = [];
        for (let y = 0; y < heightSegments; y++) {
            for (let x = 0; x < widthSegments; x++) {
                const positionIndex1 = x + y * widthPositionCount;
                const positionIndex2 = positionIndex1 + 1;
                const positionIndex3 = positionIndex1 + widthPositionCount;
                const positionIndex4 = positionIndex2 + widthPositionCount;
                const face1 = [positionIndex1, positionIndex2, positionIndex3];
                const face2 = [positionIndex2, positionIndex4, positionIndex3];
                indexArr.push(...face1, ...face2);
            }
        }
        this.setIndex(new THREE.Uint32BufferAttribute(new Uint32Array(indexArr), 1));
    }
}

/**
 * !!! 会用于 Web Worker
 * @param start
 * @param end
 * @returns
 */
/**
 * ternaryOperation(true, (v) => v, 0) => true
 *
 * ternaryOperation(false, (v) => v, 0) => 0
 * @param value
 * @param when
 * @param fallback
 * @returns
 */
function ternaryOperation(value, when, fallback) {
    return when(value) ? value : fallback;
}

function unwrapHTMLElement(ele) {
    if (ele instanceof HTMLElement) {
        return ele;
    }
    return ternaryOperation(document.body.querySelector(ele), e => e instanceof HTMLElement);
}

class BeforeLayerManager {
    layers = [];
    bbox = fullBBox;
    /**
     * 当前实际显示区域的 BBox，只会比 bbox 更小
     */
    displayBBox = fullBBox;
    z = 0;
    onUpdate;
    container;
    map;
    earthOrbitControls;
    constructor(options) {
        this.container = options.container;
        this.map = options.map;
        this.earthOrbitControls = options.earthOrbitControls;
    }
    refresh() {
        this.layers.forEach(layer => {
            layer.refresh();
        });
    }
    addLayer(layer) {
        layer.beforeLayerManager = this;
        layer.canvas.width = this.container.clientWidth;
        layer.canvas.height = this.container.clientHeight;
        layer.refresh();
        this.container.appendChild(layer.canvas);
        this.layers.push(layer);
    }
    removeLayer(layer) {
        const index = this.layers.findIndex(l => l === layer);
        this.layers.splice(index, 1);
    }
}

/**
 * 显示最上层的 layer，如 TextLayer
 */
class BaseBeforeLayer extends THREE.EventDispatcher {
    beforeLayerManager;
    zIndex = 0;
    canvas = document.createElement('canvas');
    ctx = this.canvas.getContext('2d');
    constructor() {
        super();
        this.canvas.style.position = 'absolute';
        this.canvas.style.inset = '0';
    }
}

class Map extends THREE.EventDispatcher {
    tileSize = 512;
    // 地球半径 6371km
    earthRadius = 6371;
    renderer = new THREE.WebGLRenderer();
    scene = new THREE.Scene();
    hash = false;
    earthOrbitControls;
    earthGeometry;
    earthMesh = new THREE.Mesh();
    container;
    layerManager;
    beforeLayerManager;
    tileLayer;
    disposeFuncList = [];
    /**
     * 安全区域，如果 displayBBox 超出了安全区域，则需要更新 preloadBBox
     */
    secureBBox;
    constructor(options) {
        super();
        const container = unwrapHTMLElement(options.container);
        if (!(container instanceof HTMLElement)) {
            console.error('can not find container');
            return;
        }
        container.style.position = 'relative';
        this.container = container;
        if (typeof options.hash === 'boolean')
            this.hash = options.hash;
        const pixelRatio = container.clientWidth / container.clientHeight;
        this.renderer.setSize(container.clientWidth, container.clientHeight);
        this.renderer.setPixelRatio(pixelRatio);
        container.appendChild(this.renderer.domElement);
        this.disposeFuncList.push(() => this.renderer.dispose());
        const { earthRadius, tileSize } = this;
        // let stats: Stats | undefined
        // if (options.fps) {
        //   stats = Stats()
        //   stats.dom.style.position = 'absolute'
        //   container.appendChild(stats.dom)
        // }
        this.scene.background = new THREE.Color(0x020924);
        // this.scene.fog = new THREE.Fog(0x020924, 200, 1000)
        // this.addBackground()
        // 镜头控制器
        const hashOptions = this.parseHash();
        this.earthOrbitControls = new EarthOrbitControls({
            domElement: container,
            earthRadius,
            center: options.center,
            zoom: options.zoom,
            bearing: options.bearing,
            ...hashOptions,
        });
        this.initEarthOrbitControls();
        this.earthGeometry = new EarthGeometry({
            earthRadius,
            tileSize,
            z: this.earthOrbitControls.z,
            delay: true,
        });
        this.disposeFuncList.push(() => this.earthGeometry.dispose());
        this.earthMesh.geometry = this.earthGeometry;
        this.earthMesh.userData.id = 'test';
        this.scene.add(this.earthMesh);
        this.initEarthMaterial();
        this.beforeLayerManager = new BeforeLayerManager({
            container,
            map: this,
            earthOrbitControls: this.earthOrbitControls,
        });
        // this.layerManager.canvas.style.width = '100%'
        // container.insertBefore(this.layerManager.canvas, container.childNodes[0])
        // 页面重绘动画
        const tick = () => {
            // stats?.update()
            // 更新渲染器
            this.renderer.render(this.scene, this.earthOrbitControls.camera);
            // 页面重绘时调用自身
            const id = window.requestAnimationFrame(tick);
            this.disposeFuncList.push(() => window.cancelAnimationFrame(id));
        };
        tick();
        const ro = new ResizeObserver(() => {
            const _pixelRatio = container.clientWidth / container.clientHeight;
            this.renderer.setPixelRatio(_pixelRatio);
            this.renderer.setSize(container.clientWidth, container.clientHeight);
            const { camera } = this.earthOrbitControls;
            camera.aspect = _pixelRatio;
            camera.updateProjectionMatrix();
        });
        ro.observe(container);
        this.disposeFuncList.push(() => ro.disconnect());
    }
    addBackground() {
        const { earthRadius } = this;
        // 因为是球体，需要将图片横向和竖向各翻转一次，让图片边界可以正常衔接
        const reverseRepeat = 2;
        const heightSegments = reverseRepeat * 100;
        const widthSegments = heightSegments;
        const widthPositionCount = widthSegments + 1;
        const heightPositionCount = heightSegments + 1;
        const backgroundGeometry = new THREE.SphereGeometry(earthRadius * 1000, widthSegments, heightSegments);
        const uv = [];
        for (let y = 0; y < heightPositionCount; y++) {
            for (let x = 0; x < widthPositionCount; x++) {
                const xUv = ((x * reverseRepeat) / widthSegments) % reverseRepeat;
                const yUv = ((y * reverseRepeat) / heightSegments) % reverseRepeat;
                const uvFormat = (value) => {
                    if (value > 1) {
                        value = reverseRepeat - value;
                    }
                    return value;
                };
                uv.push(uvFormat(xUv), uvFormat(yUv));
            }
        }
        backgroundGeometry.attributes.uv = new THREE.Float32BufferAttribute(new Float32Array(uv), 2);
        const texture = new THREE.TextureLoader().load('.//01-earth-splash-stars-ltr.webp');
        texture.minFilter = THREE.NearestFilter;
        const backgroundMaterial = new THREE.MeshBasicMaterial({
            map: texture,
            side: THREE.DoubleSide,
        });
        const background = new THREE.Mesh(backgroundGeometry, backgroundMaterial);
        this.scene.add(background);
    }
    updatePreloadBBox(_preloadBBox) {
        const preloadBBox = _preloadBBox ?? this.getPreloadBBox();
        const displayBBox = bbox(this.getDisplayPolygon());
        if (!isFull(preloadBBox)) {
            this.secureBBox = this.getSecureBBox();
        }
        this.earthGeometry.bbox = preloadBBox;
        this.earthGeometry.update();
        this.tileLayer.bbox = preloadBBox;
        this.tileLayer.displayBBox = displayBBox;
        this.tileLayer.z = this.earthOrbitControls.z;
        void this.tileLayer.refresh();
        this.layerManager.bbox = preloadBBox;
        this.layerManager.displayBBox = displayBBox;
        this.layerManager.z = this.earthOrbitControls.z;
        this.layerManager.updateCanvasSize(this.earthOrbitControls.getPxDeg());
        this.layerManager.refresh();
    }
    initEarthOrbitControls() {
        const onMove = throttle(() => {
            this.updateHash();
            this.beforeLayerManager.refresh();
            const secureBBox = this.secureBBox;
            const cachePreloadBBox = this.earthGeometry.bbox;
            const preloadBBox = this.getPreloadBBox();
            const displayBBox = bbox(this.getDisplayPolygon());
            if (bboxContains(cachePreloadBBox, preloadBBox)) {
                this.updatePreloadBBox();
                return;
            }
            const overflowSecureBBox = secureBBox && !bboxContains(secureBBox, displayBBox);
            if (overflowSecureBBox) {
                this.updatePreloadBBox();
            }
        }, 500);
        this.earthOrbitControls.addEventListener('move', () => {
            this.dispatchEvent({ type: 'move' });
            onMove();
        });
        this.earthOrbitControls.addEventListener('rotate', () => {
            this.dispatchEvent({ type: 'rotate' });
            onMove();
        });
        this.earthOrbitControls.addEventListener('zoom', () => {
            this.dispatchEvent({ type: 'zoom' });
            onMove();
        });
    }
    getPreloadBBox() {
        let preloadBBox = scale(bbox(this.getDisplayPolygon()), 3);
        preloadBBox = latPretreatmentBBox(preloadBBox);
        if (preloadBBox[2] - preloadBBox[0] > 180) {
            preloadBBox[0] = -180;
            preloadBBox[2] = 180;
        }
        return preloadBBox;
    }
    getSecureBBox() {
        return latPretreatmentBBox(scale(bbox(this.getDisplayPolygon()), 2));
    }
    parseHash() {
        if (this.hash && location.hash.startsWith('#')) {
            const [zoom, lng, lat, bearing] = location.hash.slice(1).split('/');
            return {
                zoom: parseFloat(zoom),
                center: [parseFloat(lng), parseFloat(lat)],
                bearing: parseFloat(bearing),
            };
        }
    }
    updateHash() {
        const { zoom, center, bearing } = this.earthOrbitControls;
        if (this.hash) {
            const arr = [floor(zoom, 2), floor(center[0], 3), floor(center[1], 3), floor(bearing, 2)];
            history.replaceState(null, '', `#${arr.join('/')}`);
        }
    }
    initEarthMaterial() {
        const { tileSize } = this;
        const materials = [];
        const backgroundMaterial = new THREE.ShaderMaterial({
            fragmentShader: `
        void main() {
          gl_FragColor = vec4(0, 0, 0, 1);
        }
      `,
        });
        materials.push(backgroundMaterial);
        const vertexShader = `
      varying vec2 vUv;
      void main() {
        vUv = uv;
        gl_Position = projectionMatrix * modelViewMatrix * vec4( position, 1.0 );
      }
    `;
        // 创建 tileMaterial
        {
            const tileLayer = new TileLayer(tileSize);
            this.tileLayer = tileLayer;
            const getTexture = () => new THREE.CanvasTexture(tileLayer.canvas);
            tileLayer.onUpdate = () => {
                uniforms.bbox.value = tileLayer.bbox;
                uniforms.canvasBBox.value = tileLayer.canvasBBox;
                uniforms.startCanvasY.value = MercatorTile.latToY(tileLayer.canvasBBox[1], 0);
                uniforms.endCanvasY.value = MercatorTile.latToY(tileLayer.canvasBBox[3], 0);
                uniforms.tile.value.dispose();
                uniforms.tile.value = getTexture();
            };
            const uniforms = {
                tile: { value: getTexture() },
                bbox: { value: fullBBox },
                canvasBBox: { value: fullBBox },
                startCanvasY: { value: MercatorTile.latToY(fullBBox[1], 0) },
                endCanvasY: { value: MercatorTile.latToY(fullBBox[3], 0) },
            };
            // 将 lat(纬度) 转化为墨卡托投影中的 y(0-1)
            const latToY = `
        float latToY(float lat) {
          if (lat >= ${MercatorTile.maxLat}) {
            return 0.0;
          }
          if (lat <= ${-MercatorTile.maxLat}) {
            return 1.0;
          }

          float sinValue = sin(radians(lat));
          float y = (0.5 - (0.25 * log((1.0 + sinValue) / (1.0 - sinValue))) / ${Math.PI});
          return y;
        }
      `;
            const fragmentShader = `
        uniform sampler2D tile;
        uniform sampler2D layers;
        uniform vec4 bbox;
        uniform vec4 canvasBBox;
        uniform float startCanvasY;
        uniform float endCanvasY;
        varying vec2 vUv;

        ${latToY}

        void main() {
          float w = bbox[0];
          float s = bbox[1];
          float e = bbox[2];
          float n = bbox[3];
          float canvasW = canvasBBox[0];
          float canvasS = canvasBBox[1];
          float canvasE = canvasBBox[2];
          float canvasN = canvasBBox[3];

          float canvasLatGap = canvasE - canvasW;
          float scaleX = (e - w) / canvasLatGap;
          float startX = (w - canvasW) / canvasLatGap;
          float x = vUv.x * scaleX + startX;

          float lat = s + vUv.y * (n - s);
          float y = (latToY(lat) - startCanvasY) / (endCanvasY - startCanvasY);

          vec2 uv = vec2(x, y);
          gl_FragColor = texture2D(tile, uv);
        }
      `;
            const tileMaterial = new THREE.ShaderMaterial({
                uniforms,
                vertexShader,
                fragmentShader,
                transparent: true,
            });
            materials.push(tileMaterial);
        }
        // 创建 layersMaterial
        {
            const layerManager = new LayerManager();
            this.layerManager = layerManager;
            this.disposeFuncList.push(() => layerManager.dispose());
            const getTexture = () => new THREE.CanvasTexture(layerManager.canvas);
            layerManager.onUpdate = () => {
                layersUniform.value.dispose();
                layersUniform.value = getTexture();
            };
            const layersUniform = {
                value: getTexture(),
            };
            const layersMaterial = new THREE.ShaderMaterial({
                uniforms: {
                    layers: layersUniform,
                },
                vertexShader,
                fragmentShader: `
          uniform sampler2D layers;
          varying vec2 vUv;
          void main() {
            gl_FragColor = texture2D(layers, vUv);
          }
        `,
                transparent: true,
            });
            materials.push(layersMaterial);
        }
        materials.forEach((material, i) => {
            this.earthGeometry.addGroup(0, Infinity, i);
            this.disposeFuncList.push(() => material.dispose());
        });
        this.earthMesh.material = materials;
        this.updatePreloadBBox();
    }
    inContainerRange(point) {
        const [x, y] = point;
        return (inRange(x, 0, this.container.clientWidth, '[]') &&
            inRange(y, 0, this.container.clientHeight, '[]'));
    }
    /**
     * 将 LngLat 转化为像素位置
     * @param lngLat
     * @param options.allowNotVisible
     * @returns
     */
    project(lngLat, options) {
        const { camera } = this.earthOrbitControls;
        const position = lngLatToVector3(lngLat, this.earthRadius);
        const vector = position.clone().project(camera);
        const w = this.container.clientWidth / 2;
        const h = this.container.clientHeight / 2;
        const x = vector.x * w + w;
        const y = -vector.y * h + h;
        if (!options?.allowNotVisible) {
            if (!this.inContainerRange([x, y])) {
                return null;
            }
            const ray = new THREE.Ray(camera.position, position.clone().sub(camera.position).normalize());
            const chordCenter = new THREE.Vector3();
            ray.closestPointToPoint(new THREE.Vector3(0, 0, 0), chordCenter);
            // 根据弦心和镜头的距离和经纬度和镜头距离的远近来判断经纬度是否被遮挡
            if (camera.position.distanceToSquared(chordCenter) < camera.position.distanceToSquared(position)) {
                return null;
            }
        }
        return [x, y];
    }
    /**
     * 将像素位置转化为 LngLat
     * 为什么不使用 THREE.Raycaster: https://github.com/mrdoob/three.js/issues/11449
     * @param point
     * @param options.allowFovLimitExceeded 是否允许该点对应的视角超出球体相切角度，如果为 true，即便超出相切角度，也会将其转化为相切角度然后返回经纬度
     * @returns
     */
    unproject(point, options) {
        if (!this.inContainerRange(point)) {
            return null;
        }
        // 将像素位置转化为以 container 中心点为原点的 xy 轴坐标
        const x = point[0] - this.container.clientWidth / 2;
        const y = -(point[1] - this.container.clientHeight / 2);
        const diagonal = hypotenuse(x, y) * 2;
        const tangentFov = getTangentFov(this.earthOrbitControls.distance, this.earthRadius);
        const fov = this.earthOrbitControls.getFov(diagonal);
        if (!options?.allowFovLimitExceeded && fov > tangentFov) {
            return null;
        }
        const centralAngle = getDisplayCentralAngle(this.earthOrbitControls.distance, this.earthRadius, fov);
        const aspect = x / y;
        let deg = Number.isNaN(aspect) ? 0 : radToDeg(Math.atan(Math.abs(x / y)));
        if (x <= 0 && y < 0) {
            deg = 180 - deg;
        }
        else if (x > 0 && y <= 0) {
            deg = 180 + deg;
        }
        else if (x > 0 && y > 0) {
            deg = 360 - deg;
        }
        const xAxis = new THREE.Vector3(-1, 0, 0).applyEuler(new THREE.Euler(0, degToRad(this.earthOrbitControls.center[0])));
        const vector3 = lngLatToVector3(this.earthOrbitControls.center, this.earthRadius)
            .applyAxisAngle(xAxis, degToRad(centralAngle / 2))
            .applyAxisAngle(lngLatToVector3(this.earthOrbitControls.center, 1), degToRad(deg - this.earthOrbitControls.bearing));
        return vector3ToLngLat(vector3);
    }
    getDisplayPolygon() {
        let bearing = this.earthOrbitControls.bearing;
        bearing %= 180;
        bearing = bearing >= 0 ? bearing : 180 + bearing;
        const diagonalDeg = radToDeg(Math.atan(this.earthOrbitControls.camera.aspect));
        let arr = [0, diagonalDeg, 90, 180 - diagonalDeg];
        const removeWhenEqBearing = (i) => {
            if (Math.round(arr[i]) === Math.round(bearing)) {
                arr.splice(i, 1);
            }
        };
        for (let i = 1; i < arr.length; i++) {
            if (bearing < arr[i]) {
                arr.splice(i, 0, bearing);
                removeWhenEqBearing(i - 1);
                removeWhenEqBearing(i + 1);
                break;
            }
            if (i === arr.length - 1) {
                arr.push(bearing);
                removeWhenEqBearing(i);
                removeWhenEqBearing(0);
                break;
            }
        }
        arr = [...arr, ...arr.map(v => v + 180)];
        let bearingIndex = arr.findIndex(v => v === bearing);
        if (this.earthOrbitControls.bearing < 0) {
            bearingIndex += arr.length / 2;
        }
        arr = [...arr.slice(bearingIndex), ...arr.slice(0, bearingIndex)];
        let lngLatArr = arr
            .map(v => rectangleIntersect(this.container.clientWidth, this.container.clientHeight, 360 - v))
            .map((point) => this.unproject(point, { allowFovLimitExceeded: true }));
        const sortLngLatArr = () => {
            const spliceIndex = lngLatArr.slice(0, -1).findIndex((v, i) => v[0] > lngLatArr[i + 1][0]);
            if (spliceIndex !== -1)
                lngLatArr = [...lngLatArr.slice(spliceIndex + 1), ...lngLatArr.slice(0, spliceIndex + 1)];
        };
        const getIntersectLat = () => {
            const start = last(lngLatArr);
            const end = [...lngLatArr[0]];
            end[0] += 360;
            const featureCollection = lineIntersect(lineString([start, end]), lineString([
                [180, -90],
                [180, 90],
            ]));
            return featureCollection.features[0]?.geometry?.coordinates?.[1];
        };
        const northPoleVisible = !!this.project([0, 90]);
        const southPoleVisible = !!this.project([0, -90]);
        if (southPoleVisible) {
            lngLatArr.reverse();
        }
        if (northPoleVisible || southPoleVisible) {
            sortLngLatArr();
            if (lngLatArr[0][0] === -180) {
                lngLatArr.push([180, lngLatArr[0][1]]);
            }
            else if (last(lngLatArr)[0] === 180) {
                lngLatArr.unshift([-180, last(lngLatArr)[1]]);
            }
            else {
                const intersectLat = getIntersectLat();
                lngLatArr.unshift([-180, intersectLat]);
                lngLatArr.push([180, intersectLat]);
            }
            const topLat = northPoleVisible ? 90 : -90;
            lngLatArr.unshift([-180, topLat]);
            lngLatArr.push([180, topLat]);
            return polygon([[...lngLatArr, lngLatArr[0]]]);
        }
        lngLatArr[0][0] = lngLatArr[lngLatArr.length / 2][0] = this.earthOrbitControls.center[0];
        lngLatArr.slice(lngLatArr.length / 2 + 1).forEach(v => {
            if (v[0] < this.earthOrbitControls.center[0]) {
                v[0] += 360;
            }
        });
        return polygon([[...lngLatArr, lngLatArr[0]]]);
    }
    addLayer(layer) {
        if (layer instanceof BaseBeforeLayer) {
            this.beforeLayerManager.addLayer(layer);
        }
        else {
            this.layerManager.addLayer(layer);
        }
    }
    removeLayer(layer) {
        if (layer instanceof BaseBeforeLayer) {
            this.beforeLayerManager.removeLayer(layer);
        }
        else {
            this.layerManager.removeLayer(layer);
        }
    }
    clearContainer() {
        this.container.childNodes.forEach(child => {
            this.container.removeChild(child);
        });
    }
    dispose() {
        this.disposeFuncList.forEach(func => func());
        this.disposeFuncList = [];
        this.clearContainer();
    }
}

function features2featureArr(f) {
    return !Array.isArray(f) && f.type === 'FeatureCollection' ? f.features : inflate(f);
}

class TextLayer extends BaseBeforeLayer {
    source;
    style;
    constructor(options) {
        super();
        this.source = options.source;
        this.style = options.style;
    }
    refresh() {
        const { canvas, ctx, source, style } = this;
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        ctx.fillStyle = '#fff';
        ctx.textAlign = 'center';
        ctx.textBaseline = 'middle';
        let fontSize = 12;
        if (style) {
            if (typeof style.fontSize === 'number') {
                fontSize = style.fontSize;
            }
            if (style.color) {
                ctx.fillStyle = style.color;
            }
        }
        ctx.font = `${fontSize}px sans-serif`;
        const bboxArr = [];
        features2featureArr(source).forEach(feature => {
            const lngLat = feature.geometry.coordinates;
            const point = this.beforeLayerManager?.map.project(lngLat);
            if (!point)
                return;
            const [x, y] = point;
            const text = feature.properties?.name;
            const textMetrics = ctx.measureText(text);
            const left = x - textMetrics.width / 2;
            const right = x + textMetrics.width / 2;
            const top = y - fontSize / 2;
            const bottom = y + fontSize / 2;
            const textBBox = [left, top, right, bottom];
            if (!bboxContains([0, 0, canvas.width, canvas.height], textBBox)) {
                return;
            }
            console.log(textBBox, bboxArr, bboxArr.some(bbox => bboxOverlap(bbox, textBBox)));
            if (!style?.overlap && bboxArr.some(bbox => bboxOverlap(bbox, textBBox))) {
                return;
            }
            ctx.fillText(text, x, y);
            bboxArr.push(textBBox);
        });
    }
}

function geoEquirectangular(options) {
    const [w, s, e, n] = options.bbox;
    const [width, height] = options.size;
    return (position) => {
        const [lng, lat] = position;
        const x = ((lng - w) / (e - w) * width);
        const y = ((n - lat) / (n - s) * height);
        return [x, y];
    };
}

class BaseLayer extends THREE.EventDispatcher {
    disposeFuncList = [];
    layerManager;
    imageBitmap;
    zIndex = 0;
    update() { }
    dispose() {
        this.disposeFuncList.forEach(func => func());
        this.disposeFuncList = [];
    }
}

class LineLayer extends BaseLayer {
    canvas = new OffscreenCanvas(1, 1);
    ctx = this.canvas.getContext('2d');
    source;
    style;
    constructor(options) {
        super();
        this.source = options.source;
        this.style = options.style;
    }
    refresh() {
        const { canvas, ctx, layerManager, source, style } = this;
        canvas.width = layerManager.canvas.width;
        canvas.height = layerManager.canvas.height;
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        const projection = geoEquirectangular({
            bbox: this.layerManager.bbox,
            size: [this.layerManager.canvas.width, this.layerManager.canvas.height],
        });
        ctx.beginPath();
        if (style) {
            if (style.lineColor) {
                this.ctx.strokeStyle = style.lineColor;
            }
            if (typeof style.lineWidth === 'number') {
                this.ctx.lineWidth = style.lineWidth;
            }
        }
        features2featureArr(source).forEach(feature => {
            let lineStringArr = [];
            if (feature.geometry.type === 'LineString') {
                lineStringArr = [feature.geometry.coordinates];
            }
            else if (feature.geometry.type === 'MultiLineString') {
                lineStringArr = feature.geometry.coordinates;
            }
            else if (feature.geometry.type === 'Polygon') {
                // TODO Polygon 支持中空（第一个环必须是外部环，其他的必须是内部环或者孔，而且内部环和外部环的走向是相反的）
                lineStringArr = feature.geometry.coordinates.slice(0, 1);
            }
            else {
                lineStringArr = feature.geometry.coordinates.map(polygonCoordinates => polygonCoordinates[0]);
            }
            lineStringArr.forEach(positions => {
                positions.forEach((position, i) => {
                    const [x, y] = projection(position);
                    if (i === 0) {
                        ctx.moveTo(x, y);
                    }
                    else {
                        ctx.lineTo(x, y);
                    }
                });
            });
        });
        ctx.stroke();
        this.imageBitmap = canvas.transferToImageBitmap();
    }
}

class FillLayer extends BaseLayer {
    canvas = new OffscreenCanvas(1, 1);
    ctx = this.canvas.getContext('2d');
    source;
    style;
    constructor(options) {
        super();
        this.source = options.source;
        this.style = options.style;
    }
    refresh() {
        const { canvas, ctx, layerManager, source, style } = this;
        canvas.width = layerManager.canvas.width;
        canvas.height = layerManager.canvas.height;
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        const projection = geoEquirectangular({
            bbox: this.layerManager.bbox,
            size: [this.layerManager.canvas.width, this.layerManager.canvas.height],
        });
        ctx.beginPath();
        if (style) {
            if (style.fillColor) {
                ctx.fillStyle = style.fillColor;
            }
        }
        features2featureArr(source).forEach(feature => {
            let coordinates = [];
            if (feature.geometry.type === 'MultiPolygon') {
                coordinates = feature.geometry.coordinates.map(polygonCoordinates => polygonCoordinates[0]);
            }
            else {
                coordinates = feature.geometry.coordinates.slice(0, 1);
            }
            coordinates.forEach(positions => {
                positions.forEach((position, i) => {
                    const [x, y] = projection(position);
                    if (i === 0) {
                        ctx.moveTo(x, y);
                    }
                    else {
                        ctx.lineTo(x, y);
                    }
                });
            });
        });
        ctx.fill();
        this.imageBitmap = canvas.transferToImageBitmap();
    }
}

export { FillLayer, LineLayer, Map, TextLayer };
